"""
bengrn base module.
"""

import pandas as pd
import urllib.request
import os.path
import json

from grnndata import GRNAnnData, from_adata_and_longform, from_scope_loomfile, utils
from anndata import AnnData, concat
from typing import Optional, Union
from .tools import GENIE3
import numpy as np
import bionty as bt
from anndata.utils import make_index_unique

from arboreto.algo import grnboost2

# from pyscenic.utils import modules_from_adjacencies
# from pyscenic.prune import prune2df, df2regulons
# from pyscenic.aucell import aucell
# issue here of using an older version of numpy calling np.object instead of np.object_

from ctxcore.rnkdb import FeatherRankingDatabase as RankingDatabase

import scipy.stats
from sklearn.metrics import precision_recall_curve, auc, PrecisionRecallDisplay
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import train_test_split
from sklearn.linear_model import RidgeClassifier

import gseapy as gp

import tqdm
import logging

import matplotlib.pyplot as plt

from scipy.sparse import csr_matrix, csc_matrix

import scanpy as sc

# example constant variable
NAME = "bengrn"

FILEDIR = os.path.dirname(os.path.realpath(__file__))


# Define a function to calculate precision and recall
def precision_recall(true_con, grn_con):
    """
    Calculate precision and recall from the true and predicted connections.

    Args:
        true_con (np.array): The true connections.
        grn_con (np.array): The predicted connections.

    Returns:
        float: The precision value.
        float: The recall value.
    """
    tp = len(true_con & grn_con)
    fp = len(grn_con - true_con)
    fn = len(true_con - grn_con)

    precision = tp / (tp + fp)
    recall = tp / (tp + fn)

    return precision, recall


class BenGRN:
    def __init__(
        self,
        grn: GRNAnnData,
        full_dataset: Optional[AnnData] = None,
        doplot: bool = True,
        do_auc: bool = True,
    ):
        """
        Initializes the BenGRN class.

        Args:
            grn (GRNAnnData): The Gene Regulatory Network data.
            full_dataset (Optional[AnnData]): The full dataset, defaults to None.
            doplot (bool): Whether to plot the results, defaults to True.
            do_auc (bool): Whether to calculate the Area Under the Precision-Recall Curve, defaults to True.
        """
        self.grn = grn
        self.full_dataset = full_dataset
        self.doplot = doplot
        self.do_auc = do_auc

    def get_self_metrics(self):
        print("I'm getting my own metrics")
        pass

        # recap N protein complexes

        # gene pairs reg correlation with high string score vs none

        # recap well-known relationships

        #

    def compare_to(
        self,
        other: Optional[GRNAnnData] = None,
        to: str = "collectri",
        organism: str = "human",
    ):
        """
        compare_to compares the GRN to another GRN.

        Args:
            other (Optional[GRNAnnData], optional): The other GRN to compare to. Defaults to None.
                If not given can use a default GRN from the 'to' argument.
            to (str, optional): The name of the other GRN to compare to. Defaults to "collectri".
                If 'other' is given, this argument is ignored.
            organism (str, optional): The organism of the GRN to compare to. Defaults to "human".

        Returns:
            dict: The metrics of the comparison.
        """
        if other is None:
            if self.doplot:
                print("loading GT, ", to)
            gt = get_GT_db(name=to, organism=organism)
            # gt = gt[gt.type != "post_translational"]
            varnames = set(gt.iloc[:, :2].values.flatten())
            intersection = varnames & set(self.grn.var["symbol"].tolist())
            loc = self.grn.var["symbol"].isin(intersection)
            adj = self.grn.varp["GRN"][:, loc][loc, :]
            genes = self.grn.var.loc[loc, "symbol"].tolist()

            da = np.zeros(adj.shape, dtype=float)
            for i, j in gt.iloc[:, :2].values:
                if i in genes and j in genes:
                    da[genes.index(i), genes.index(j)] = 1
            if self.doplot:
                print("intersection of {} genes".format(len(intersection)))
                print("intersection pct:", len(intersection) / len(self.grn.grn.index))
        else:
            elems = other.var[other.grn.sum(1) != 0].index.tolist()
            da = other.get(self.grn.var.index.tolist()).get(elems).targets
            if da.shape[1] < 5:
                print("da is very small: ", da.shape[1])
            # da = da.iloc[6:]
            adj = self.grn.grn.loc[da.index.values, da.columns.values].values
            da = da.values

        return compute_pr(
            adj,
            da,
            doplot=self.doplot,
            do_auc=self.do_auc,
        )

    def scprint_benchmark(self, elems=["Central", "Regulators", "Targets"]):
        """
        scprint_benchmark full benchmarks of the GRN as in the scPRINT paper.

        It will apply first an enrichment analysis over the [elems] of the GRN looking for TF enrichment and cell type marker gene enrichment
        It will then apply an enrichment over each TF in the GRN for their targets in ENCODE.
        Finaly, it will compare it to the OmniPath database GRN using precision recall type metrics.

        Args:
            elems (list, optional): The genes in the GRN, to benchmark. Defaults to ["Central", "Regulators", "Targets"].
                It corresponds to different views of the network.

        Returns:
            dict: The metrics of the benchmark.
        """
        print("base enrichment")
        metrics = {}
        for elem in elems:
            if elem == "Central" and (self.grn.varp["GRN"] != 0).sum() > 100_000_000:
                print("too many genes for central computation")
                continue
            res = utils.enrichment(
                self.grn,
                of=elem,
                gene_sets=[
                    {"TFs": [i.split(".")[0] for i in utils.TF]},
                    utils.file_dir + "/celltype.gmt",
                ],
                doplot=False,
                maxsize=2000,
                top_k=10,
            )
            if res is None:
                continue
            if (
                len(res.res2d[(res.res2d["FDR q-val"] < 0.1) & (res.res2d["NES"] > 1)])
                > 0
            ):
                metrics.update(
                    {
                        "enriched_terms_"
                        + elem: res.res2d[
                            (res.res2d["FDR q-val"] < 0.1) & (res.res2d["NES"] > 1)
                        ].Term.tolist()
                    }
                )
            if self.doplot:
                try:
                    _ = res.plot(terms="0__TFs")
                    plt.show()
                except KeyError:
                    pass
            istrue = metrics.get("TF_enr", False)
            istrue = istrue or (
                res.res2d.loc[res.res2d.Term == "0__TFs", "FDR q-val"].iloc[0] < 0.1
            )
            metrics.update({"TF_enr": istrue})
        if self.doplot:
            print("_________________________________________")
            print("TF specific enrichment")
        with open(FILEDIR + "/../data/tfchip_data.json", "r") as file:
            tfchip = json.load(file)
        TFinchip = {i: i.split(" ")[0] for i in tfchip.keys()}
        res = {}
        i, j = 0, 0
        previous_level = logging.root.manager.disable
        logging.disable(logging.WARNING)
        for k, v in TFinchip.items():
            if v not in self.grn.grn.columns:
                continue
            # print(k)
            j += 1
            test = self.grn.grn.T.loc[[v]].sort_values(by=v, ascending=False).T
            if len(set(test.index) & set(tfchip[k])) == 0:
                continue
            if test.iloc[:, 0].sum() == 0:
                continue
            try:
                pre_res = gp.prerank(
                    rnk=test,
                    gene_sets=[{v: tfchip[k]}],
                    background=self.grn.var.index.tolist(),
                    min_size=1,
                    max_size=4000,
                    permutation_num=1000,
                )
            except IndexError:
                continue
            val = (
                pre_res.res2d[
                    (pre_res.res2d["FDR q-val"] < 0.05) & (pre_res.res2d["NES"] > 1)
                ]
                .sort_values(by=["NES"], ascending=False)
                .drop(columns=["Name"])
            )
            if len(val.Term.tolist()) > 0:
                # print("found! ", val.Term.tolist()[0])
                # print(pre_res.res2d["NES"])
                # print(pre_res.res2d["FDR q-val"])
                # print("\n")
                i += 1
            else:
                pass
                # print("no sig...")
            res[k] = pre_res.res2d
            # print(val.Term.tolist()[:2])
        logging.disable(previous_level)
        j = j if j != 0 else 1
        if self.doplot:
            print("found some significant results for ", i * 100 / j, "% TFs\n")
            print("_________________________________________")
        metrics.update({"significant_enriched_TFtargets": i * 100 / j})
        metrics.update(self.compare_to(to="omnipath", organism="human"))
        return metrics


def train_classifier(
    grn: GRNAnnData,
    gt: str = "omnipath",
    other: Optional[GRNAnnData] = None,
    use_col: str = "symbol",
    train_size: float = 0.2,
    doplot: bool = True,
    class_weight: dict = {1: 200, 0: 1},
    max_iter: int = 1_000,
    C: float = 1.0,
    return_full: bool = True,
    shuffle: bool = False,
    **kwargs,
):
    """
    train_classifier trains a classifier to generate a GRN that maps to the ground truth.

    Uses a RidgeClassifier to select the best combination of networks to predict the ground truth.
    It is used for the head classification part in the scPRINT paper.

    Args:
        grn (GRNAnnData): The Gene Regulatory Network data.
        gt (str, optional): The name of the ground truth database to use. Defaults to "omnipath".
        other (GRNAnnData, optional): Another GRN to compare against. Defaults to None.
        use_col (str, optional): The column name to use for gene symbols. Defaults to "symbol".
        train_size (float, optional): The proportion of the dataset to include in the train split. Defaults to 0.2.
        doplot (bool, optional): Whether to plot the results. Defaults to True.
        class_weight (dict, optional): Weights associated with classes in the form {class_label: weight}. Defaults to {1: 200, 0: 1}.
        max_iter (int, optional): Maximum number of iterations for the classifier. Defaults to 1_000.
        C (float, optional): Regularization strength; must be a positive float. Defaults to 1.0.
        return_full (bool, optional): Whether to return the full classifier object. Defaults to True.
        shuffle (bool, optional): Whether to shuffle the data before splitting. Defaults to False.

    Returns:
        (GRNAnnData, dict, RidgeClassifier): The Gene Regulatory Network data, the metrics of the classifier, and the classifier object.
    """
    if other is not None:
        elems = other.var[other.grn.sum(1) != 0].index.tolist()
        sub = other.get(grn.var[use_col].tolist()).get(elems).targets
        if sub.shape[1] < 5:
            print("sub is very small: ", sub.shape[1])
        genes = grn.var[use_col].tolist()
        args = np.argsort(genes)
        genes = np.array(genes)[args]
        adj = grn.varp["GRN"][args, :, :][:, args, :][np.isin(genes, sub.index.values)][
            :, np.isin(genes, sub.columns.values)
        ]
        print("pred shape", adj.shape)
        da = sub.values
    else:
        gt = get_GT_db(name=gt)
        varnames = set(gt.iloc[:, :2].values.flatten())
        intersection = varnames & set(grn.var[use_col].tolist())
        loc = grn.var[use_col].isin(intersection)
        adj = grn.varp["GRN"][:, loc, :][loc, :, :]
        genes = grn.var.loc[loc][use_col].tolist()

        da = np.zeros((len(genes), len(genes)), dtype=float)
        for i, j in gt.iloc[:, :2].values:
            if i in genes and j in genes:
                da[genes.index(i), genes.index(j)] = 1

    print("true elem", int(da.sum()), "...")
    da = da.flatten()
    adj = adj.reshape(-1, adj.shape[-1])

    X_train, X_test, y_train, y_test = train_test_split(
        adj, da, random_state=0, train_size=train_size, shuffle=shuffle
    )
    print("doing classification....")

    clf = RidgeClassifier(
        alpha=C,
        fit_intercept=False,
        class_weight=class_weight,
        # solver="saga",
        max_iter=max_iter,
        positive=True,
    )
    # clf = LogisticRegression(
    #    penalty="l1",
    #    C=C,
    #    solver="saga",
    #    class_weight=class_weight,
    #    max_iter=max_iter,
    #    n_jobs=8,
    #    fit_intercept=False,
    #    verbose=10,
    #    **kwargs,
    # )
    clf.fit(X_train, y_train)
    pred = clf.predict(X_test)
    # epr = compute_epr(clf, X_test, y_test)
    metrics = {
        "used_heads": (clf.coef_ != 0).sum(),
        "precision": (pred[y_test == 1] == 1).sum() / (pred == 1).sum(),
        "random_precision": y_test.sum() / len(y_test),
        "recall": (pred[y_test == 1] == 1).sum() / y_test.sum(),
        "predicted_true": pred.sum(),
        "number_of_true": y_test.sum(),
        # "epr": epr,
    }
    if doplot:
        print("metrics", metrics)
        PrecisionRecallDisplay.from_estimator(
            clf, X_test, y_test, plot_chance_level=True
        )
        plt.show()
    if return_full:
        adj = grn.varp["GRN"]
        grn.varp["classified"] = clf.predict(adj.reshape(-1, adj.shape[-1])).reshape(
            len(grn.var), len(grn.var)
        )
    return grn, metrics, clf


def get_scenicplus(
    filepath=FILEDIR + "/../data/10xPBMC_homo_scenicplus_genebased_scope.loom",
):
    """
    This function retrieves a loomx scenicplus data from a given file path and loads it as a GrnnData

    Args:
        filepath : str, optional
            The path to the scenicplus data file.
            Default is FILEDIR + "/../data/10xPBMC_homo_scenicplus_genebased_scope.loom".

    Raises:
        FileNotFoundError: If the file at the given path does not exist.

    Returns:
        GrnAnnData: The scenicplus data from the given file as a grnndata object
    """
    if not os.path.exists(filepath):
        raise FileNotFoundError(
            f"The file {filepath} does not exist. You likely need to download \
                this or another loomxfile from the scope website"
        )

    return from_scope_loomfile(filepath)


def get_sroy_gt(
    get: str = "main", join: str = "outer", species: str = "human", gt: str = "full"
) -> GrnAnnData:
    """
    This function retrieves the ground truth data from the McCall et al.'s paper.

    Args:
        get (str): The specific dataset to retrieve. Options include "main", "liu", and "chen".
        join (str, optional): The type of join to be performed when concatenating the data. Default is "outer".
        species (str, optional): The species of the dataset. Default is "human".
        gt (str, optional): The type of ground truth data to retrieve. Options include "full", "chip", and "ko". Default is "full".

    Returns:
        GrnAnnData: The ground truth data as a grnndata object
    """
    if species == "human":
        if gt == "full":
            df = pd.read_csv(
                FILEDIR + "/../data/GroundTruth/stone_and_sroy/hESC_ground_truth.tsv",
                sep="\t",
                header=None,
            )
        elif gt == "chip":
            df = pd.read_csv(
                FILEDIR
                + "/../data/GroundTruth/stone_and_sroy/gold_standards/hESC/hESC_chipunion.txt",
                sep="\t",
                header=None,
            )
        elif gt == "ko":
            df = pd.read_csv(
                FILEDIR
                + "/../data/GroundTruth/stone_and_sroy/gold_standards/hESC/hESC_KDUnion.txt",
                sep="\t",
                header=None,
            )
        if get == "liu":
            adata = AnnData(
                (
                    2
                    ** pd.read_csv(
                        FILEDIR
                        + "/../data/GroundTruth/stone_and_sroy/scRNA/liu_rna_filtered_log2.tsv.gz",
                        sep="\t",
                    )
                )
                - 1
            ).T
        elif get == "chen":
            adata = AnnData(
                (
                    2
                    ** pd.read_csv(
                        FILEDIR
                        + "/../data/GroundTruth/stone_and_sroy/scRNA/chen_rna_filtered_log2.tsv.gz",
                        sep="\t",
                    )
                )
                - 1
            ).T
        elif get == "han":
            adata = AnnData(
                unnormalize(
                    pd.read_csv(
                        FILEDIR
                        + "/../data/GroundTruth/stone_and_sroy/scRNA/human_han_GSE107552.csv.gz",
                    )
                    .set_index("Cell")
                    .T,
                    is_root=True,
                )
            )
        elif get == "mine":
            # https://www.ebi.ac.uk/gxa/sc/experiments/E-GEOD-36552/downloads
            adata = sc.read_mtx(
                FILEDIR
                + "/../data/GroundTruth/stone_and_sroy/scRNA/E-GEOD-36552.aggregated_filtered_counts.mtx"
            ).T
            col = pd.read_csv(
                FILEDIR
                + "/../data/GroundTruth/stone_and_sroy/scRNA/E-GEOD-36552.aggregated_filtered_counts.mtx_rows",
                header=None,
                sep="\t",
            )
            adata.var.index = col[0]
            genesdf = load_genes()
            intersect_genes = set(adata.var.index).intersection(set(genesdf.index))
            adata = adata[:, list(intersect_genes)]
            genesdf = genesdf.loc[adata.var.index]
            adata.var["ensembl_id"] = adata.var.index
            adata.var.index = make_index_unique(genesdf["symbol"].astype(str))
        else:
            raise ValueError("get must be one of 'liu', 'chen', 'han', or 'mine'")
        adata.obs["organism_ontology_term_id"] = "NCBITaxon:9606"
    elif species == "mouse":
        if gt == "full":
            df = pd.read_csv(
                FILEDIR + "/../data/GroundTruth/stone_and_sroy/mESC_ground_truth.tsv",
                sep="\t",
                header=None,
            )
        elif gt == "chip":
            df = pd.read_csv(
                FILEDIR
                + "/../data/GroundTruth/stone_and_sroy/gold_standards/mESC/mESC_chipunion.txt",
                sep="\t",
                header=None,
            )
        elif gt == "ko":
            df = pd.read_csv(
                FILEDIR
                + "/../data/GroundTruth/stone_and_sroy/gold_standards/mESC/mESC_KDUnion.txt",
                sep="\t",
                header=None,
            )
        if get == "duren":
            adata = AnnData(
                (
                    2
                    ** pd.read_csv(
                        FILEDIR
                        + "/../data/GroundTruth/stone_and_sroy/scRNA/duren_rna_filtered_log2.tsv.gz",
                        sep="\t",
                    )
                )
                - 1
            ).T
        elif get == "semrau":
            adata = AnnData(
                (
                    2
                    ** pd.read_csv(
                        FILEDIR
                        + "/../data/GroundTruth/stone_and_sroy/scRNA/semrau_rna_filtered_log2.tsv.gz",
                        sep="\t",
                    )
                )
                - 1
            ).T
        elif get == "tran":
            adata = AnnData(
                unnormalize(
                    pd.read_csv(
                        FILEDIR
                        + "/../data/GroundTruth/stone_and_sroy/scRNA/mouse_tran_A2S.csv.gz",
                    )
                    .set_index("Cell")
                    .T,
                    is_root=True,
                )
            )
        elif get == "zhao":
            adata = AnnData(
                unnormalize(
                    pd.read_csv(
                        FILEDIR
                        + "/../data/GroundTruth/stone_and_sroy/scRNA/mouse_zhao_GSE114952.csv.gz",
                    )
                    .set_index("Cell")
                    .T,
                    is_root=True,
                )
            )
        else:
            raise ValueError("get must be one of 'duren', 'semrau', 'tran', or 'zhao'")
        adata.obs["organism_ontology_term_id"] = "NCBITaxon:10090"
    return from_adata_and_longform(adata, df)


def get_perturb_gt(
    url_bh: str = "https://plus.figshare.com/ndownloader/files/38349308",
    filename_bh: str = FILEDIR + "/../data/BH-corrected.csv.gz",
    url_adata: str = "https://plus.figshare.com/ndownloader/files/35773219",
    filename_adata: str = FILEDIR + "/../data/ess_perturb_sc.h5ad",
):
    """
    get_perturb_gt retrieves the genome wide perturb seq ground truth data.

    Args:
        url_bh (str, optional): The URL to download the BH-corrected data. Defaults to "https://plus.figshare.com/ndownloader/files/38349308".
        filename_bh (str, optional): The local filename to save the BH-corrected data. Defaults to FILEDIR + "/../data/BH-corrected.csv.gz".
        url_adata (str, optional): The URL to download the single-cell perturbation data. Defaults to "https://plus.figshare.com/ndownloader/files/35773219".
        filename_adata (str, optional): The local filename to save the single-cell perturbation data. Defaults to FILEDIR + "/../data/ess_perturb_sc.h5ad".

    Returns:
        GRNAnnData: The Gene Regulatory Network data as a GRNAnnData object.
    """
    if not os.path.exists(filename_bh):
        urllib.request.urlretrieve(url_bh, filename_bh)
    pert = pd.read_csv(filename_bh)
    pert = pert.set_index("Unnamed: 0").T
    pert.index = [i.split("_")[-1] for i in pert.index]
    pert = pert[~pert.index.duplicated(keep="first")].T
    pert = pert < 0.05
    adata_sc = sc.read(
        filename_adata,
        backup_url=url_adata,
    )
    adata_sc = adata_sc[adata_sc.obs.gene_id == "non-targeting"]
    adata_sc[:, adata_sc.var.index.isin(set(pert.index) | set(pert.columns))]
    adata_sc.obs["organism_ontology_term_id"] = "NCBITaxon:9606"
    adata_sc = adata_sc[:, adata_sc.var.sort_index().index]

    pert = pert.loc[pert.index.isin(adata_sc.var.index)].loc[
        :, pert.columns.isin(adata_sc.var.index)
    ]

    missing_indices = list(set(adata_sc.var.index) - set(pert.index))
    pert = pert.reindex(pert.index.union(missing_indices), fill_value=False)
    missing_indices = list(set(adata_sc.var.index) - set(pert.columns))
    pert = pert.reindex(columns=pert.columns.union(missing_indices), fill_value=False)

    pert = pert.loc[adata_sc.var.index].loc[:, adata_sc.var.index]
    return GRNAnnData(
        X=csr_matrix(adata_sc.X.toarray()),
        var=adata_sc.var,
        obs=adata_sc.obs,
        grn=csr_matrix(pert.values),
    )


def compute_scenic(adata, data_dir=FILEDIR + "/../data"):
    """
    This function computes the SCENIC algorithm on the given dataset.

    Args:
        adata (AnnData): The annotated data matrix of shape n_obs x n_vars. Rows correspond to cells and columns to genes.
        data_dir (str, optional): The directory where the data files will be stored. Defaults to FILEDIR + "/../data".

    Returns:
        GRNAnnData: The Gene Regulatory Network data.
    """
    os.makedirs(data_dir, exist_ok=True)

    url1 = "https://resources.aertslab.org/cistarget/motif2tf/motifs-v10nr_clust-nr.hgnc-m0.001-o0.0.tbl"
    url2 = "https://resources.aertslab.org/cistarget/databases/homo_sapiens/hg38/refseq_r80/mc_v10_clust/gene_based/hg38_10kbp_up_10kbp_down_full_tx_v10_clust.genes_vs_motifs.rankings.feather"

    file1 = os.path.join(data_dir, os.path.basename(url1))
    file2 = os.path.join(data_dir, os.path.basename(url2))

    if not os.path.exists(file1):
        urllib.request.urlretrieve(url1, file1)
    if not os.path.exists(file2):
        urllib.request.urlretrieve(url2, file2)
    dbs = RankingDatabase(fname=file2, name="genes_10kb_human_rank_v10")
    adjacencies = grnboost2(
        expression_data=adata.to_df(), tf_names=utils.TF, verbose=True
    )
    modules = list(modules_from_adjacencies(adjacencies, adata.to_df()))

    df = prune2df([dbs], modules, file1)
    regulons = df2regulons(df)
    # auc_mtx = aucell(adata.to_df(), regulons, num_workers=8)
    # sns.clustermap(auc_mtx, figsize=(12,12))

    # compute the grn matrix
    var_names = adata.var_names.tolist()
    da = np.zeros((len(var_names), len(var_names)), dtype=float)
    for reg in regulons:
        i = var_names.index(reg.transcription_factor)
        for k, v in reg.gene2weight.items():
            da[i, var_names.index(k)] = v
    # create the grndata
    grn = GRNAnnData(adata, grn=da)

    da = np.zeros((len(var_names), len(var_names)), dtype=float)
    for reg in modules:
        i = var_names.index(reg.transcription_factor)
        for k, v in reg.gene2weight.items():
            da[i, var_names.index(k)] = v
    grn.varp["modules"] = da
    grn.var["TFs"] = [True if i in utils.TF else False for i in grn.var_names]
    return grn


def compute_genie3(
    adata: AnnData, nthreads: int = 30, ntrees: int = 100, **kwargs
) -> GRNAnnData:
    """
    This function computes the GENIE3 algorithm on the given data.

    Args:
        adata (AnnData): The annotated data matrix of shape n_obs x n_vars. Rows correspond to cells and columns to genes.
        nthreads (int, optional): The number of threads to use for computation. Defaults to 30.
        ntrees (int, optional): The number of trees to use for the Random Forests. Defaults to 100.
        **kwargs: Additional arguments to pass to the GENIE3 function.

    Returns:
        GRNAnnData: The Gene Regulatory Network data computed using the GENIE3 algorithm.
    """
    mat = np.asarray(adata.X.todense())
    names = adata.var_names[mat.sum(0) > 0].tolist()
    var = adata.var[mat.sum(0) > 0]
    mat = mat[:, mat.sum(0) > 0]
    VIM = GENIE3(mat, gene_names=names, nthreads=nthreads, ntrees=ntrees, **kwargs)
    grn = GRNAnnData(grn=VIM, X=mat, var=var, obs=adata.obs)
    grn.var_names = grn.var["symbol"]
    grn.var["TFs"] = [True if i in utils.TF else False for i in grn.var_names]
    return grn


def get_GT_db(
    name: str = "collectri", organism: str = "human", split_complexes: bool = True
) -> pd.DataFrame:
    """
    use_prior_network loads a prior GRN from a list of available networks.

    Args:
        name (str, optional): name of the network to load. Defaults to "collectri".
        organism (str, optional): organism to load the network for. Defaults to "human".
        split_complexes (bool, optional): whether to split complexes into individual genes. Defaults to True.

    Returns:
        pd.DataFrame: The prior GRN as a pandas DataFrame.

    Raises:
        ValueError: if the provided name is not amongst the available names.
    """
    if name == "tflink":
        TFLINK = "https://cdn.netbiol.org/tflink/download_files/TFLink_Homo_sapiens_interactions_All_simpleFormat_v1.0.tsv.gz"
        net = pd_load_cached(TFLINK)
        net = net.rename(columns={"Name.TF": "source", "Name.Target": "target"})
    elif name == "htftarget":
        HTFTARGET = "http://bioinfo.life.hust.edu.cn/static/hTFtarget/file_download/tf-target-infomation.txt"
        net = pd_load_cached(HTFTARGET)
        net = net.rename(columns={"TF": "source"})
    elif name == "collectri":
        import decoupler as dc

        net = dc.get_collectri(organism=organism, split_complexes=split_complexes).drop(
            columns=["PMID"]
        )
    elif name == "dorothea":
        import decoupler as dc

        net = dc.get_dorothea(organism=organism)
    elif name == "omnipath":
        if not os.path.exists(FILEDIR + "/../data/omnipath.parquet"):
            from omnipath.interactions import AllInteractions
            from omnipath.requests import Annotations

            interactions = AllInteractions()
            net = interactions.get(exclude=["small_molecule", "lncrna_mrna"])
            hgnc = Annotations.get(resources="HGNC")
            rename = {v.uniprot: v.genesymbol for k, v in hgnc.iterrows()}
            net.source = net.source.replace(rename)
            net.target = net.target.replace(rename)
            net.to_parquet(FILEDIR + "/../data/omnipath.parquet")
        else:
            net = pd.read_parquet(FILEDIR + "/../data/omnipath.parquet")
    else:
        raise ValueError(f"provided name: '{name}' is not amongst the available names.")
    # varnames = list(set(net.iloc[:, :2].values.flatten()))
    # adata = AnnData(var=varnames)
    # adata.var_names = varnames
    # grn = from_adata_and_longform(adata, net, has_weight=True)
    # return grn
    return net


def pd_load_cached(url, loc="/tmp/", cache=True, **kwargs):
    # Check if the file exists, if not, download it
    loc += url.split("/")[-1]
    if not os.path.isfile(loc) or not cache:
        urllib.request.urlretrieve(url, loc)
    # Load the data from the file
    return pd.read_csv(loc, **kwargs)


def compute_pr(
    grn: np.array,
    true: np.array,
    do_auc: bool = True,
    doplot: bool = True,
):
    """
    compute_pr computes the precision and recall metrics for the given GRN and true matrix.

    Args:
        grn (np.array): The Gene Regulatory Network matrix, where each element represents the strength of the regulatory relationship between genes.
        true (np.array): The ground truth matrix, where each element indicates the presence (1) or absence (0) of a regulatory relationship.
        do_auc (bool, optional): Whether to compute the Area Under the Precision-Recall Curve (AUPRC). Defaults to True.
        doplot (bool, optional): Whether to plot the precision and recall metrics. Defaults to True.

    Raises:
        ValueError: If the shape of the GRN and the true matrix do not match.

    Returns:
        dict: A dictionary containing precision, recall, and random precision metrics.
    """
    if grn.shape != true.shape:
        raise ValueError("The shape of the GRN and the true matrix do not match.")
    metrics = {}
    true = true.astype(bool)
    tot = (grn.shape[0] * grn.shape[1]) - grn.shape[0]
    precision = (grn[true] != 0).sum() / (grn != 0).sum()
    recall = (grn[true] != 0).sum() / true.sum()
    rand_prec = true.sum() / tot

    if doplot:
        print(
            "precision: ",
            precision,
            "\nrecall: ",
            recall,
            "\nrandom precision:",
            rand_prec,
        )
    metrics.update(
        {
            "precision": precision,
            "recall": recall,
            "rand_precision": rand_prec,
        }
    )
    # Initialize lists to store precision and recall values
    precision_list = [precision]
    recall_list = [recall]
    # Define the thresholds to vary
    thresholds = np.append(
        np.linspace(0, 1, 101)[:-2], np.log10(np.logspace(0.99, 1, 30))
    )
    thresholds = np.quantile(grn, thresholds)
    # Calculate precision and recall for each threshold
    if do_auc:
        for threshold in tqdm.tqdm(thresholds[1:]):
            precision = (grn[true] > threshold).sum() / (grn > threshold).sum()
            recall = (grn[true] > threshold).sum() / true.sum()
            precision_list.append(precision)
            recall_list.append(recall)
        # Calculate AUPRC by integrating the precision-recall curve
        if 1.0 not in recall_list:
            precision_list.insert(0, rand_prec)
            recall_list.insert(0, recall_list[0])
            precision_list.insert(0, rand_prec)
            recall_list.insert(0, 1.0)
        precision_list = np.nan_to_num(np.array(precision_list))
        recall_list = np.nan_to_num(np.array(recall_list))
        auprc = -np.trapz(precision_list, recall_list)
        metrics["auprc"] = auprc

        # Compute Average Precision (AP) manually
        sorted_indices = np.argsort(-grn.flatten())
        sorted_true = true.flatten()[sorted_indices]

        tp_cumsum = np.cumsum(sorted_true)
        fp_cumsum = np.cumsum(~sorted_true)

        precision_at_k = tp_cumsum / (tp_cumsum + fp_cumsum)
        recall_at_k = tp_cumsum / true.sum()

        ap = np.sum(precision_at_k[1:] * np.diff(recall_at_k))
        metrics["ap"] = ap
        if doplot:
            print("Average Precision (AP): ", ap)
        if doplot:
            print("Area Under Precision-Recall Curve (AUPRC): ", auprc)

    # compute EPR
    # get the indices of the topK highest values in "grn"
    if isinstance(grn, csr_matrix):
        grn = grn.toarray()
    if isinstance(grn, csc_matrix):
        grn = grn.toarray()
    indices = np.argpartition(grn.flatten(), -int(true.sum()))[-int(true.sum()) :]
    # Compute the odds ratio
    true_positive = true[np.unravel_index(indices, true.shape)].sum()
    false_positive = true.sum() - true_positive
    # this is normal as we compute on the same number of pred_pos as true_pos
    false_negative = true.sum() - true_positive
    true_negative = tot - true_positive - false_positive - false_negative
    # Avoid division by zero
    # this is a debugger line
    if true_negative == 0 or false_positive == 0:
        odds_ratio = float("inf")
    else:
        odds_ratio = (true_positive * true_negative) / (false_positive * false_negative)

    metrics.update({"epr": odds_ratio})
    if doplot:
        print("EPR:", odds_ratio)
        plt.figure(figsize=(10, 8))
        plt.plot(
            recall_list,
            precision_list,
            marker=".",
            linestyle="-",
            color="b",
            label="p-r",
        )
        plt.plot(
            [recall_list[0], recall_list[-1]],
            [rand_prec, rand_prec],
            linestyle="--",
            color="r",
            label="Random Precision",
        )
        plt.legend(loc="lower left")
        plt.title("Precision-Recall Curve")
        plt.xlabel("Recall")
        plt.ylabel("Precision")
        plt.xscale("log")
        plt.grid(True)
        plt.show()
    return metrics


def compute_epr(clf, X_test: np.ndarray, y_test: np.ndarray) -> float:
    """
    compute_epr computes the Expected Precision Recall (EPR) metric for the given classifier, test data, and true labels.

    Args:
        clf (sklearn.base.ClassifierMixin): The classifier to evaluate.
        X_test (numpy.ndarray): The test data features.
        y_test (numpy.ndarray): The true labels for the test data.

    Returns:
        float: The computed Expected Precision Recall (EPR) metric.
    """
    prb = clf.predict_proba(X_test)[:, 1]

    K = sum(y_test)
    # get only the top-K elems from prb
    pred = np.zeros(prb.shape)
    pred[np.argsort(prb)[-int(K) :]] = 1

    true_positive = np.sum(pred[y_test == 1] == 1)
    false_positive = np.sum(pred[y_test == 0] == 1)
    false_negative = np.sum(pred[y_test == 1] == 0)
    true_negative = np.sum(pred[y_test == 0] == 0)
    odds_ratio = (true_positive * true_negative) / (false_positive * false_negative)
    return odds_ratio


def unnormalize(df, is_root=False):
    # rescale (logp1 or sqroot)
    df = df**2 if is_root else (2**df) - 1
    r = np.array([i[i != 0].min() for k, i in df.iterrows()])
    df = (df.T / r).T.round()
    return df


def load_genes(organisms: Union[str, list] = "NCBITaxon:9606"):  # "NCBITaxon:10090",
    """
    load_genes loads the genes for the given organisms.

    Args:
        organisms (Union[str, list], optional): The organism(s) to load genes for. Can be a single organism string or a list of organism strings. Defaults to "NCBITaxon:9606".

    Returns:
        pd.DataFrame: A DataFrame containing gene information for the specified organisms, including columns for gene symbols, mitochondrial genes, ribosomal genes, hemoglobin genes, and organism.
    """
    organismdf = []
    if type(organisms) == str:
        organisms = [organisms]
    for organism in organisms:
        genesdf = bt.Gene.filter(
            organism_id=bt.Organism.filter(ontology_id=organism).first().id
        ).df()
        genesdf = genesdf[~genesdf["public_source_id"].isna()]
        genesdf = genesdf.drop_duplicates(subset="ensembl_gene_id")
        genesdf = genesdf.set_index("ensembl_gene_id").sort_index()
        # mitochondrial genes
        genesdf["mt"] = genesdf.symbol.astype(str).str.startswith("MT-")
        # ribosomal genes
        genesdf["ribo"] = genesdf.symbol.astype(str).str.startswith(("RPS", "RPL"))
        # hemoglobin genes.
        genesdf["hb"] = genesdf.symbol.astype(str).str.contains(("^HB[^(P)]"))
        genesdf["organism"] = organism
        organismdf.append(genesdf)
    return pd.concat(organismdf)



from sklearn.tree import BaseDecisionTree
from sklearn.ensemble import RandomForestRegressor, ExtraTreesRegressor
import time
from operator import itemgetter
from multiprocessing import Pool
from typing import Optional, Union
from numpy import ndarray, array, transpose, std, zeros, ndenumerate, random


def compute_feature_importances(estimator):
    if isinstance(estimator, BaseDecisionTree):
        return estimator.tree_.compute_feature_importances(normalize=False)
    else:
        importances = [
            e.tree_.compute_feature_importances(normalize=False)
            for e in estimator.estimators_
        ]
        importances = array(importances)
        return sum(importances, axis=0) / len(estimator)


def get_link_list(
    VIM: ndarray,
    gene_names: Optional[list[str]] = None,
    regulators: Union[list[str], str] = "all",
    maxcount: Union[int, str] = "all",
    file_name: Optional[str] = None,
):
    """Gets the ranked list of (directed) regulatory links.

    Args
        VIM (np.array): Array as returned by the function GENIE3(), in which the element (i,j) is the score of the edge directed from the i-th gene to the j-th gene.
        gene_names (list[str] optional): List of length p, where p is the number of rows/columns in VIM, containing the names of the genes. The i-th item of gene_names must correspond to the i-th row/column of VIM. When the gene names are not provided, the i-th gene is named Gi. Default is None.
        regulators (list[str], optional): List containing the names of the candidate regulators. When a list of regulators is provided, the names of all the genes must be provided (in gene_names), and the returned list contains only edges directed from the candidate regulators. When regulators is set to 'all', any gene can be a candidate regulator. Default is 'all'.
        maxcount (Union[str, int], optional): Writes only the first maxcount regulatory links of the ranked list. When maxcount is set to 'all', all the regulatory links are written. Default is 'all'.
        file_name (str, optional): Writes the ranked list of regulatory links to the file file_name. Default is None.

    Returns
        pd.Dataframe: The list of regulatory links, ordered according to the edge score. Auto-regulations do not appear in the list. Regulatory links with a score equal to zero are randomly permuted. In the ranked list of edges, each line has format:
            regulator   target gene     score of edge
    """

    # Check input arguments
    if not isinstance(VIM, ndarray):
        raise ValueError("VIM must be a square array")
    elif VIM.shape[0] != VIM.shape[1]:
        raise ValueError("VIM must be a square array")

    ngenes = VIM.shape[0]

    if gene_names is not None:
        if not isinstance(gene_names, (list, tuple)):
            raise ValueError("input argument gene_names must be a list of gene names")
        elif len(gene_names) != ngenes:
            raise ValueError(
                "input argument gene_names must be a list of length p, where p is the number of columns/genes in the expression data"
            )

    if regulators != "all":
        if not isinstance(regulators, (list, tuple)):
            raise ValueError("input argument regulators must be a list of gene names")

        if gene_names is None:
            raise ValueError(
                "the gene names must be specified (in input argument gene_names)"
            )
        else:
            sIntersection = set(gene_names).intersection(set(regulators))
            if not sIntersection:
                raise ValueError(
                    "The genes must contain at least one candidate regulator"
                )

    if maxcount != "all" and not isinstance(maxcount, int):
        raise ValueError('input argument maxcount must be "all" or a positive integer')

    if file_name is not None and not isinstance(file_name, str):
        raise ValueError("input argument file_name must be a string")

    # Get the indices of the candidate regulators
    if regulators == "all":
        input_idx = range(ngenes)
    else:
        input_idx = [i for i, gene in enumerate(gene_names) if gene in regulators]

    # Get the non-ranked list of regulatory links
    vInter = [
        (i, j, score) for (i, j), score in ndenumerate(VIM) if i in input_idx and i != j
    ]

    # Rank the list according to the weights of the edges
    vInter_sort = sorted(vInter, key=itemgetter(2), reverse=True)
    nInter = len(vInter_sort)

    # Random permutation of edges with score equal to 0
    flag = 1
    i = 0
    while flag and i < nInter:
        (TF_idx, target_idx, score) = vInter_sort[i]
        if score == 0:
            flag = 0
        else:
            i += 1

    if not flag:
        items_perm = vInter_sort[i:]
        items_perm = random.permutation(items_perm)
        vInter_sort[i:] = items_perm

    # Write the ranked list of edges
    nToWrite = nInter
    if isinstance(maxcount, int) and maxcount >= 0 and maxcount < nInter:
        nToWrite = maxcount

    if file_name:
        outfile = open(file_name, "w")

        if gene_names is not None:
            for i in range(nToWrite):
                (TF_idx, target_idx, score) = vInter_sort[i]
                TF_idx = int(TF_idx)
                target_idx = int(target_idx)
                outfile.write(
                    "%s\t%s\t%.6f\n"
                    % (gene_names[TF_idx], gene_names[target_idx], score)
                )
        else:
            for i in range(nToWrite):
                (TF_idx, target_idx, score) = vInter_sort[i]
                TF_idx = int(TF_idx)
                target_idx = int(target_idx)
                outfile.write("G%d\tG%d\t%.6f\n" % (TF_idx + 1, target_idx + 1, score))

        outfile.close()

    else:
        if gene_names is not None:
            for i in range(nToWrite):
                (TF_idx, target_idx, score) = vInter_sort[i]
                TF_idx = int(TF_idx)
                target_idx = int(target_idx)
                print(
                    "%s\t%s\t%.6f" % (gene_names[TF_idx], gene_names[target_idx], score)
                )
        else:
            for i in range(nToWrite):
                (TF_idx, target_idx, score) = vInter_sort[i]
                TF_idx = int(TF_idx)
                target_idx = int(target_idx)
                print("G%d\tG%d\t%.6f" % (TF_idx + 1, target_idx + 1, score))


def GENIE3(
    expr_data: ndarray,
    gene_names: Optional[list[str]] = None,
    regulators: Union[list[str], str] = "all",
    tree_method: str = "RF",
    K: Union[str, int] = "sqrt",
    ntrees: int = 1000,
    nthreads: int = 1,
):
    """
    GENIE3 Computation of tree-based scores for all putative regulatory links.

    Args:
        expr_data (numpy.ndarray): Array containing gene expression values. Each row corresponds to a condition and each column corresponds to a gene.
        gene_names (list[str], optional): List of length p, where p is the number of columns in expr_data, containing the names of the genes. The i-th item of gene_names must correspond to the i-th column of expr_data. Defaults to None.
        regulators (list[str], optional): List containing the names of the candidate regulators. When a list of regulators is provided, the names of all the genes must be provided (in gene_names). When regulators is set to 'all', any gene can be a candidate regulator. Defaults to 'all'.
        tree_method (str, optional): Specifies which tree-based procedure is used: either Random Forest ('RF') or Extra-Trees ('ET'). Defaults to 'RF'.
        K (str or int, optional): Specifies the number of selected attributes at each node of one tree: either the square root of the number of candidate regulators ('sqrt'), the total number of candidate regulators ('all'), or any positive integer. Defaults to 'sqrt'.
        ntrees (int, optional): Specifies the number of trees grown in an ensemble. Defaults to 1000.
        nthreads (int, optional): Number of threads used for parallel computing. Defaults to 1.

    Returns:
        numpy.ndarray: An array in which the element (i,j) is the score of the edge directed from the i-th gene to the j-th gene. All diagonal elements are set to zero (auto-regulations are not considered). When a list of candidate regulators is provided, the scores of all the edges directed from a gene that is not a candidate regulator are set to zero.
    """
    time_start = time.time()

    # Check input arguments
    if not isinstance(expr_data, ndarray):
        raise ValueError(
            "expr_data must be an array in which each row corresponds to a condition/sample and each column corresponds to a gene"
        )

    ngenes = expr_data.shape[1]

    if gene_names is not None:
        if not isinstance(gene_names, (list, tuple)):
            raise ValueError("input argument gene_names must be a list of gene names")
        elif len(gene_names) != ngenes:
            raise ValueError(
                "input argument gene_names must be a list of length p, where p is the number of columns/genes in the expr_data"
            )

    if regulators != "all":
        if not isinstance(regulators, (list, tuple)):
            raise ValueError("input argument regulators must be a list of gene names")

        if gene_names is None:
            raise ValueError(
                "the gene names must be specified (in input argument gene_names)"
            )
        else:
            sIntersection = set(gene_names).intersection(set(regulators))
            if not sIntersection:
                raise ValueError(
                    "the genes must contain at least one candidate regulator"
                )

    if tree_method != "RF" and tree_method != "ET":
        raise ValueError(
            'input argument tree_method must be "RF" (Random Forests) or "ET" (Extra-Trees)'
        )

    if K != "sqrt" and K != "all" and not isinstance(K, int):
        raise ValueError(
            'input argument K must be "sqrt", "all" or a stricly positive integer'
        )

    if isinstance(K, int) and K <= 0:
        raise ValueError(
            'input argument K must be "sqrt", "all" or a stricly positive integer'
        )

    if not isinstance(ntrees, int):
        raise ValueError("input argument ntrees must be a stricly positive integer")
    elif ntrees <= 0:
        raise ValueError("input argument ntrees must be a stricly positive integer")

    if not isinstance(nthreads, int):
        raise ValueError("input argument nthreads must be a stricly positive integer")
    elif nthreads <= 0:
        raise ValueError("input argument nthreads must be a stricly positive integer")

    print("Tree method: " + str(tree_method))
    print("K: " + str(K))
    print("Number of trees: " + str(ntrees))
    print("\n")

    # Get the indices of the candidate regulators
    if regulators == "all":
        input_idx = list(range(ngenes))
    else:
        input_idx = [i for i, gene in enumerate(gene_names) if gene in regulators]

    # Learn an ensemble of trees for each target gene, and compute scores for candidate regulators
    VIM = zeros((ngenes, ngenes))

    if nthreads > 1:
        print("running jobs on %d threads" % nthreads)

        input_data = list()
        for i in range(ngenes):
            input_data.append([expr_data, i, input_idx, tree_method, K, ntrees])

        pool = Pool(nthreads)
        alloutput = pool.map(wr_GENIE3_single, input_data)

        for i, vi in alloutput:
            VIM[i, :] = vi

    else:
        print("running single threaded jobs")
        for i in range(ngenes):
            print("Gene %d/%d..." % (i + 1, ngenes))

            vi = GENIE3_single(expr_data, i, input_idx, tree_method, K, ntrees)
            VIM[i, :] = vi

    VIM = transpose(VIM)

    time_end = time.time()
    print("Elapsed time: %.2f seconds" % (time_end - time_start))

    return VIM


def wr_GENIE3_single(args):
    return [
        args[1],
        GENIE3_single(args[0], args[1], args[2], args[3], args[4], args[5]),
    ]


def GENIE3_single(expr_data, output_idx, input_idx, tree_method, K, ntrees):
    ngenes = expr_data.shape[1]

    # Expression of target gene
    output = expr_data[:, output_idx]

    # Normalize output data
    output = output / std(output)

    # Remove target gene from candidate regulators
    input_idx = input_idx[:]
    if output_idx in input_idx:
        input_idx.remove(output_idx)

    expr_data_input = expr_data[:, input_idx]

    # Parameter K of the tree-based method
    if (K == "all") or (isinstance(K, int) and K >= len(input_idx)):
        max_features = "auto"
    else:
        max_features = K

    if tree_method == "RF":
        treeEstimator = RandomForestRegressor(
            n_estimators=ntrees, max_features=max_features
        )
    elif tree_method == "ET":
        treeEstimator = ExtraTreesRegressor(
            n_estimators=ntrees, max_features=max_features
        )

    # Learn ensemble of trees
    treeEstimator.fit(expr_data_input, output)

    # Compute importance scores
    feature_importances = compute_feature_importances(treeEstimator)
    vi = zeros(ngenes)
    vi[input_idx] = feature_importances

    return vi



import os
import yaml
import numpy as np
import pandas as pd
from pathlib import Path


def heavisideThreshold(value):
    """
    Decides whether the basal parameter omega_0 should
    be 1 (basal expression) or -1 (basal repression)
    """
    if int(value) == 1:
        return 1
    elif int(value) == 0:
        return -1


def loadParameterValues():
    """
    Checks for valid parameters in parameters.yaml file
    """
    with open(
        str(Path(__file__).parent.absolute()) + "/parameters.yaml", "r"
    ) as parameterfile:
        parameterDefaults = yaml.safe_load(parameterfile)

    requiredPars = [
        "mRNATranscription",
        "mRNADegradation",
        "proteinTranslation",
        "proteinDegradation",
        "heavisideSigma",
        "signalingTimescale",
        "hillCoefficient",
        "interactionStrength",
    ]
    boolodeDefaults = {
        "mRNATranscription": 20.0,
        "mRNADegradation": 10.0,
        "proteinTranslation": 10.0,
        "proteinDegradation": 1.0,
        "heavisideSigma": 10.0,
        "signalingTimescale": 5.0,
        "hillCoefficient": 10.0,
        "interactionStrength": 1.0,
    }
    kineticParameterDefaults = {}
    for rp in requiredPars:
        if rp not in parameterDefaults:
            print("%s is missing from the config file. Using default value.")
        kineticParameterDefaults[rp] = parameterDefaults.get(rp, boolodeDefaults[rp])

    # Max level checks
    x_max = (
        kineticParameterDefaults["mRNATranscription"]
        / kineticParameterDefaults["mRNADegradation"]
    )
    y_max = x_max * (
        kineticParameterDefaults["proteinTranslation"]
        / kineticParameterDefaults["proteinDegradation"]
    )

    if x_max != 2.0:
        print("Warning: max(mRNA) != 2.0")
    if y_max != 20.0:
        print("Warning: max(protein) != 20.0")

    return kineticParameterDefaults


def getRegulatorsInRule(rule, species, inputs):
    """
    Helper function to tokenize a rule into regulators.
    Returns three lists of regulator names.
    1. allreg is the list of all valid regulators
    2. regulatorySpecies are other model variables that are regulators
    3. inputreg are the regulators that are model inputs
    """
    rhs = rule
    rhs = rhs.replace("(", " ")
    rhs = rhs.replace(")", " ")
    tokens = rhs.split(" ")

    allreg = set([t for t in tokens if (t in species or t in inputs)])
    regulatorySpecies = set([t for t in tokens if t in species])
    inputreg = set([t for t in tokens if t in inputs])

    return (allreg, regulatorySpecies, inputreg)


def getSaneNval(size, lo=1.0, hi=10.0, mu=2.0, sig=2.0, identicalPars=False):
    """
    Generates a gaussian random number which is
    bounded by `lo` and `hi`

    :param size: number of random numbers to generate
    :type size: int
    :param lo: lower bound of sample range
    :type lo: float
    :param hi: upper bound of sample range
    :type hi: float
    :param mu: Mean of Gaussian distribution to be sampled from
    :type mu: float
    :param sigma: Standard deviation of Gausssian distribution to be sampled from
    :type sigma: float
    :param identicalPars: Flag to sample single value and return a list of identical values
    :type identicalPars: bool
    :returns:
        - K: list of sampled values
    """

    if identicalPars:
        k = np.random.normal(mu, sig)
        while k < lo or k > hi:
            k = np.random.normal(mu, sig)
        K = [k for i in range(size)]
    else:
        K = []
        for _ in range(size):
            k = np.random.normal(mu, sig)
            while k < lo or k > hi:
                k = np.random.normal(mu, sig)
            K.append(k)
    return K


def minmaxnorm(X):
    """Scales the values in X

    :param X: Input list of values to be scaled
    :type X: list
    :returns:
        - N : list of values scaled between min and max values in input list
    """
    mix = min(X)
    mx = max(X)
    N = [(x - mix) / (mx - mix) for x in X]
    return N


def normalizeData(P):
    """
    Calls minmaxnorm() for each time series

    :returns:
        Pnorm : list of scaled values
    """
    Pnorm = []
    for i in range(np.shape(P)[1]):
        Pnorm.append(minmaxnorm(P[:, i]))
    return Pnorm


def normalizeExp(DF):
    """
    Calls minmaxnorm() for each gene across all experiments

    :param DF: Dataframe containing gene expression values as rows, time points as columns
    :type DF: pd.DataFrame
    :returns:
        newDF : DataFrame with values of DF scaled between min and max of DF
    """
    genes = DF.index
    newDF = DF.copy()
    Pnorm = []
    for g in genes:
        P = DF.loc[g].values
        newDF.loc[g] = minmaxnorm(P)

    return newDF


def get_ss(P):
    """
    Return the final time point of the simulated time course

    :param P: 2-D array containing time course of the model
    :type P: np.array
    :returns:
        - ss : np.array containing the last entry in each time series list
    """
    ss = []
    ss = [p for p in P[-1, :]]
    return ss


def generateInputFiles(
    resultDF, BoolDF, withoutRules, parameterInputsDF, tmax, numcells, outPrefix=""
):
    """
    Generates input files required from the Beeline pipeline

    :param resultDF: The simulation output, rows are genes, columns are "cells" or timepoints
    :type resultDF: pandas DataFrame
    :param outputfilenames: List of filenames generated containing individual time courses
    :type outputfilenames: list
    :param BoolDF: Dataframe containing rules
    :type BoolDF: pandas DataFrame
    :param withoutrules: List of nodes in input file without rules
    :type withoutrules: list
    :param parameterInputsPath: Specifies if there are any inputs to the model
    :type parameterInputsPath: str
    :param outPrefix: Prefix specifying target directory
    :type outPrefix: str (Optional)
    """

    print("1. refNetwork")
    refnet = []
    genes = set(BoolDF["Gene"].values)
    genes = genes.difference(set(withoutRules))
    inputs = withoutRules

    for g in genes:
        row = BoolDF[BoolDF["Gene"] == g]
        rhs = list(row["Rule"].values)[0]
        rule = list(row["Rule"].values)[0]
        rhs = rhs.replace("(", " ")
        rhs = rhs.replace(")", " ")
        tokens = rhs.split(" ")
        if len(withoutRules) == 0:
            inputs = []
            avoidthese = ["and", "or", "not", ""]
        else:
            avoidthese = list(withoutRules)
            avoidthese.extend(["and", "or", "not", ""])

        regulators = [
            t for t in tokens if (t in genes or t in inputs) if t not in avoidthese
        ]
        if "not" in tokens:
            whereisnot = tokens.index("not")
        else:
            whereisnot = None
        for r in regulators:
            if whereisnot is None:
                ty = "+"
            else:
                if type(whereisnot) is int:
                    whereisnot = [whereisnot]

                if tokens.index(r) < whereisnot[0]:
                    ty = "+"
                else:
                    ty = "-"
            # Regulator is Gene1 and Target is Gene2
            refnet.append({"Gene2": g, "Gene1": r, "Type": ty})
    refNetDF = pd.DataFrame(refnet)
    refNetDF.drop_duplicates(inplace=True)
    refNetDF.to_csv(str(outPrefix) + "/refNetwork.csv", sep=",", index=False)

    # PseudoTime.csv
    print("2. PseudoTime.csv")
    cellID = list(resultDF.columns)
    time = [float(c.split("_")[1].replace("-", ".")) for c in cellID]
    experiment = [int(c.split("_")[0].split("E")[1]) for c in cellID]
    pseudotime = minmaxnorm(time)
    cellID = [c.replace("-", "_") for c in cellID]

    PseudoTimeDict = {
        "Cell ID": cellID,
        "PseudoTime": pseudotime,
        "Time": time,
        "Experiment": experiment,
    }
    PseudoTimeDF = pd.DataFrame(PseudoTimeDict)
    PseudoTimeDF.to_csv(str(outPrefix) + "/PseudoTime.csv", sep=",", index=False)
    PseudoTimeDF.index = PseudoTimeDF["Cell ID"]

    # ExpressionData.csv
    if len(resultDF.columns) < 1e3:
        print("3. ExpressionData.csv")
        columns = list(resultDF.columns)
        columns = [c.replace("-", "_") for c in columns]
        resultDF.columns = columns
        if parameterInputsDF is not None:
            resultDF = resultDF.drop(withoutRules, axis=0)
        resultDF.to_csv(str(outPrefix) + "/ExpressionData.csv", sep=",")
    else:
        print(
            "Dataset too large."
            "\nSampling %d cells, one from each simulated trajectory." % numcells
        )
        times = np.random.choice([i for i in range(1, tmax * 100)], numcells)
        expdf = pd.DataFrame(
            columns=["E" + str(i) + "_" + str(times[i]) for i in range(numcells)],
            index=resultDF.index,
        )
        for c in expdf.columns:
            expdf[c] = resultDF[c]
        expdf.to_csv(str(outPrefix) + "/ExpressionData.csv", sep=",")


def sampleTimeSeries(
    num_timepoints,
    expnum,
    tspan,
    P,
    varmapper,
    timeIndex,
    genelist,
    proteinlist,
    header,
    writeProtein=False,
):
    """
    Returns pandas DataFrame with columns corresponding to
    time points and rows corresponding to genes
    """
    revvarmapper = {v: k for k, v in varmapper.items()}
    experimentTimePoints = [h for h in header if "E" + str(expnum) in h]
    rnaIndex = [i for i in range(len(varmapper.keys())) if "x_" in varmapper[i]]
    sampleDict = {}

    if writeProtein:
        # Write protein and mRNA to file
        for ri in varmapper.keys():
            sampleDict[varmapper[ri]] = {
                h: P[ri][ti] for h, ti in zip(experimentTimePoints, timeIndex)
            }
    else:
        # Default, only mRNA
        if len(proteinlist) == 0:
            for ri in rnaIndex:
                sampleDict[varmapper[ri]] = {
                    h: P[ri][ti] for h, ti in zip(experimentTimePoints, timeIndex)
                }
        else:
            speciesoi = [revvarmapper["p_" + p] for p in proteinlist]
            speciesoi.extend([revvarmapper["x_" + g] for g in genelist])
            result = pd.DataFrame(index=pd.Index([varmapper[i] for i in speciesoi]))

            for si in speciesoi:
                sampleDict[varmapper[si]] = {
                    h: P[si][ti] for h, ti in zip(experimentTimePoints, timeIndex)
                }

    sampleDF = pd.DataFrame(sampleDict)
    return sampleDF


def sampleCellFromTraj(
    cellid,
    tspan,
    P,
    varmapper,
    sampleAt,
    genelist,
    proteinlist,
    header,
    writeProtein=False,
):
    """
    Returns pandas DataFrame with columns corresponding to
    time points and rows corresponding to genes
    """
    revvarmapper = {v: k for k, v in varmapper.items()}
    rnaIndex = [i for i in range(len(varmapper.keys())) if "x_" in varmapper[i]]
    sampleDict = {}
    timepoint = int(header[cellid].split("_")[1])
    if writeProtein:
        # Write protein and mRNA to file
        for ri in varmapper.keys():
            sampleDict[varmapper[ri]] = {header[cellid]: P[ri][timepoint]}
    else:
        # Default, only mRNA
        if len(proteinlist) == 0:
            for ri in rnaIndex:
                sampleDict[varmapper[ri]] = {header[cellid]: P[ri][timepoint]}
        else:
            speciesoi = [revvarmapper["p_" + p] for p in proteinlist]
            speciesoi.extend([revvarmapper["x_" + g] for g in genelist])
            result = pd.DataFrame(index=pd.Index([varmapper[i] for i in speciesoi]))
            for si in speciesoi:
                sampleDict[varmapper[si]] = {header[cellid]: P[ri][timepoint]}

    sampleDF = pd.DataFrame(sampleDict)
    return sampleDF


def checkValidInputPath(path):
    """
    Returns dataframe of file at path.
    If path is not valid, returns empty dataframe.
    """
    df = pd.DataFrame()
    if path.is_file():
        df = pd.read_csv(path, sep="\t", engine="python")
    return df


def checkValidModelDefinitionPath(path, name):
    # Check if model defintion file exists
    if not os.path.isfile(path):
        print("Error in definition of job " + name)
        print("Please specify path to Boolean model")
        return False
    return True



#!/usr/bin/env python
# coding: utf-8
import os
import ast
import time
import numpy as np
import pandas as pd
from tqdm import tqdm
from pathlib import Path
from sklearn.cluster import KMeans
from importlib.machinery import SourceFileLoader
import multiprocessing as mp

# local imports
from . import utils
from .model_generator import GenerateModel
from . import simulator

np.seterr(all="raise")


def Experiment(
    mg, Model, tspan, settings, icsDF, writeProtein=False, normalizeTrajectory=False
):
    """
    Carry out an `in-silico` experiment. This function takes as input
    an ODE model defined as a python function and carries out stochastic
    simulations. BoolODE defines a _cell_ as a single time point from
    a simulated time course. Thus, in order to obtain 50 single cells,
    BoolODE carries out 50 simulations, which are stored in ./simulations/.
    Further, if it is known that the resulting single cell dataset will
    exhibit multiple trajectories, the user can specify  the number of clusters in
    `nClusters`; BoolODE will then cluster the entire simulation, such that each
    simulated trajectory possesess a cluster ID.

    :param mg: Model details obtained by instantiating an object of GenerateModel
    :type mg: BoolODE.GenerateModel
    :param Model: Function defining ODE model
    :type Model: function
    :param tspan: Array of time points
    :type tspan: ndarray
    :param settings: The job settings dictionary
    :type settings: dict
    :param icsDF: Dataframe specifying initial condition for simulation
    :type icsDF: pandas DataFrame
    :param writeProtein: Bool specifying if the protein values should be written to file. Default = False
    :type writeProtein: bool
    :param normalizeTrajectory: Bool specifying if the gene expression values should be scaled between 0 and 1.
    :type normalizeTrajectory: bool
    """
    ####################
    allParameters = dict(mg.ModelSpec["pars"])
    parNames = sorted(list(allParameters.keys()))
    ## Use default parameters
    pars = [mg.ModelSpec["pars"][k] for k in parNames]
    ####################
    rnaIndex = [i for i in range(len(mg.varmapper.keys())) if "x_" in mg.varmapper[i]]
    revvarmapper = {v: k for k, v in mg.varmapper.items()}
    proteinIndex = [
        i for i in range(len(mg.varmapper.keys())) if "p_" in mg.varmapper[i]
    ]

    y0 = [mg.ModelSpec["ics"][mg.varmapper[i]] for i in range(len(mg.varmapper.keys()))]
    ss = np.zeros(len(mg.varmapper.keys()))

    for i, k in mg.varmapper.items():
        if "x_" in k:
            ss[i] = 1.0
        elif "p_" in k:
            if k.replace("p_", "") in mg.proteinlist:
                # Seting them to the threshold
                # causes them to drop to 0 rapidly
                # TODO: try setting to threshold < v < y_max
                ss[i] = 20.0

    if not icsDF.empty:
        icsspec = icsDF.loc[0]
        genes = ast.literal_eval(icsspec["Genes"])
        values = ast.literal_eval(icsspec["Values"])
        icsmap = {g: v for g, v in zip(genes, values)}
        for i, k in mg.varmapper.items():
            for p in mg.proteinlist:
                if p in icsmap.keys():
                    ss[revvarmapper["p_" + p]] = icsmap[p]
                else:
                    ss[revvarmapper["p_" + p]] = 0.01
            for g in mg.genelist:
                if g in icsmap.keys():
                    ss[revvarmapper["x_" + g]] = icsmap[g]
                else:
                    ss[revvarmapper["x_" + g]] = 0.01

    if len(mg.proteinlist) == 0:
        result = pd.DataFrame(index=pd.Index([mg.varmapper[i] for i in rnaIndex]))
    else:
        speciesoi = [revvarmapper["p_" + p] for p in proteinlist]
        speciesoi.extend([revvarmapper["x_" + g] for g in mg.genelist])
        result = pd.DataFrame(index=pd.Index([mg.varmapper[i] for i in speciesoi]))

    # Index of every possible time point. Sample from this list
    startat = 0
    timeIndex = [i for i in range(startat, len(tspan))]

    ## Construct dictionary of arguments to be passed
    ## to simulateAndSample(), done in parallel
    outPrefix = str(settings["outprefix"])
    argdict = {}
    argdict["mg"] = mg
    argdict["allParameters"] = allParameters
    argdict["parNames"] = parNames
    argdict["Model"] = Model
    argdict["tspan"] = tspan
    argdict["varmapper"] = mg.varmapper
    argdict["timeIndex"] = timeIndex
    argdict["genelist"] = mg.genelist
    argdict["proteinlist"] = mg.proteinlist
    argdict["writeProtein"] = writeProtein
    argdict["outPrefix"] = outPrefix
    argdict["sampleCells"] = settings[
        "sample_cells"
    ]  # TODO consider removing this option
    argdict["pars"] = pars
    argdict["ss"] = ss
    argdict["ModelSpec"] = mg.ModelSpec
    argdict["rnaIndex"] = rnaIndex
    argdict["proteinIndex"] = proteinIndex
    argdict["revvarmapper"] = revvarmapper
    argdict["x_max"] = mg.kineticParameterDefaults["x_max"]

    if settings["sample_cells"]:
        # pre-define the time points from which a cell will be sampled
        # per simulation
        sampleAt = np.random.choice(timeIndex, size=settings["num_cells"])
        header = [
            "E" + str(cellid) + "_" + str(time)
            for cellid, time in zip(range(settings["num_cells"]), sampleAt)
        ]

        argdict["header"] = header
    else:
        # initialize dictionary to hold raveled values, used to cluster
        # This will be useful later.
        groupedDict = {}

    simfilepath = Path(outPrefix, "./simulations/")
    if not os.path.exists(simfilepath):
        print(simfilepath, "does not exist, creating it...")
        os.makedirs(simfilepath)
    print("Starting simulations")
    start = time.time()

    if settings["doParallel"]:
        with mp.Pool() as pool:
            jobs = []
            for cellid in range(settings["num_cells"]):
                cell_args = dict(argdict, seed=cellid, cellid=cellid)
                job = pool.apply_async(simulateAndSample, args=(cell_args,))
                jobs.append(job)

            for job in jobs:
                job.wait()
    else:
        for cellid in tqdm(range(settings["num_cells"])):
            argdict["seed"] = cellid
            argdict["cellid"] = cellid
            simulateAndSample(argdict)

    print("Simulations took %0.3f s" % (time.time() - start))
    frames = []
    print("starting to concat files")
    start = time.time()

    for cellid in tqdm(range(settings["num_cells"])):
        if settings["sample_cells"]:
            df = pd.read_csv(
                outPrefix + "/simulations/E" + str(cellid) + "-cell.csv", index_col=0
            )
            df = df.sort_index()
        else:
            df = pd.read_csv(
                outPrefix + "/simulations/E" + str(cellid) + ".csv", index_col=0
            )
            df = df.sort_index()
            groupedDict["E" + str(cellid)] = df.values.ravel()
        frames.append(df.T)
    stop = time.time()
    print("Concating files took %.2f s" % (stop - start))
    result = pd.concat(frames, axis=0)
    result = result.T
    indices = result.index
    newindices = [i.replace("x_", "") for i in indices]
    result.index = pd.Index(newindices)

    if settings["nClusters"] > 1:
        ## Carry out k-means clustering to identify which
        ## trajectory a simulation belongs to
        print("Starting k-means clustering")
        groupedDF = pd.DataFrame.from_dict(groupedDict)
        print("Clustering simulations...")
        start = time.time()
        # Find clusters in the experiments
        clusterLabels = (
            KMeans(n_clusters=settings["nClusters"], n_jobs=8)
            .fit(groupedDF.T.values)
            .labels_
        )
        print("Clustering took %0.3fs" % (time.time() - start))
        clusterDF = pd.DataFrame(
            data=clusterLabels, index=groupedDF.columns, columns=["cl"]
        )
        clusterDF.to_csv(outPrefix + "/ClusterIds.csv")
    else:
        print("Requested nClusters=1, not performing k-means clustering")
    ##################################################

    return result


def startRun(settings):
    """
    Start a simulation run. Loads model file, starts an Experiment(),
    and generates the appropriate input files
    """
    validInput = utils.checkValidModelDefinitionPath(
        settings["modelpath"], settings["name"]
    )
    startfull = time.time()

    outdir = settings["outprefix"]
    if not os.path.exists(outdir):
        print(outdir, "does not exist, creating it...")
        os.makedirs(outdir)

    ##########################################
    ## Read advanced model specification files
    ## If these are not specified, the dataFrame objects
    ## are left empty
    parameterInputsDF = utils.checkValidInputPath(settings["parameter_inputs_path"])
    parameterSetDF = utils.checkValidInputPath(settings["parameter_set"])
    icsDF = utils.checkValidInputPath(settings["icsPath"])
    interactionStrengthDF = utils.checkValidInputPath(settings["interaction_strengths"])

    speciesTypeDF = utils.checkValidInputPath(settings["species_type"])
    ##########################################

    # Simulator settings
    tmax = settings["simulation_time"]
    integration_step_size = settings["integration_step_size"]
    tspan = np.linspace(0, tmax, int(tmax / integration_step_size))

    # Generate the ODE model from the specified boolean model
    mg = GenerateModel(
        settings, parameterInputsDF, parameterSetDF, interactionStrengthDF
    )
    genesDict = {}

    # Load the ODE model file
    model = SourceFileLoader("model", mg.path_to_ode_model.as_posix()).load_module()

    ## Function call - do the in silico experiment
    resultDF = Experiment(
        mg,
        model.Model,
        tspan,
        settings,
        icsDF,
        writeProtein=settings["writeProtein"],
        normalizeTrajectory=settings["normalizeTrajectory"],
    )

    # Write simulation output. Creates ground truth files.
    print("Generating input files for pipline...")
    start = time.time()
    utils.generateInputFiles(
        resultDF,
        mg.df,
        mg.withoutRules,
        parameterInputsDF,
        tmax,
        settings["num_cells"],
        outPrefix=settings["outprefix"],
    )
    print("Input file generation took %0.2f s" % (time.time() - start))
    print("BoolODE.py took %0.2fs" % (time.time() - startfull))


def simulateAndSample(argdict):
    """
    Handles parallelization of ODE simulations.
    Calls the simulator with simulation settings.
    """
    mg = argdict["mg"]
    allParameters = argdict["allParameters"]
    parNames = argdict["parNames"]
    Model = argdict["Model"]
    tspan = argdict["tspan"]
    varmapper = argdict["varmapper"]
    timeIndex = argdict["timeIndex"]
    genelist = argdict["genelist"]
    proteinlist = argdict["proteinlist"]
    writeProtein = argdict["writeProtein"]
    cellid = argdict["cellid"]
    outPrefix = argdict["outPrefix"]
    sampleCells = argdict["sampleCells"]
    ss = argdict["ss"]
    ModelSpec = argdict["ModelSpec"]
    rnaIndex = argdict["rnaIndex"]
    proteinIndex = argdict["proteinIndex"]
    genelist = argdict["genelist"]
    proteinlist = argdict["proteinlist"]
    revvarmapper = argdict["revvarmapper"]
    seed = argdict["seed"]
    pars = argdict["pars"]
    x_max = argdict["x_max"]

    # Retained for debugging
    isStochastic = True

    if sampleCells:
        header = argdict["header"]

    pars = {}
    for k, v in allParameters.items():
        pars[k] = v
    pars = [pars[k] for k in parNames]

    ## Boolean to check if a simulation is going to a
    ## 0 steady state, with all genes/proteins dying out
    retry = True
    trys = 0
    ## timepoints
    tps = [i for i in range(1, len(tspan))]
    ## gene ids
    gid = [i for i, n in varmapper.items() if "x_" in n]
    outPrefix = outPrefix + "/simulations/"
    while retry:
        seed += 1000
        y0_exp = simulator.getInitialCondition(
            ss,
            ModelSpec,
            rnaIndex,
            proteinIndex,
            genelist,
            proteinlist,
            varmapper,
            revvarmapper,
        )

        P = simulator.simulateModel(Model, y0_exp, pars, isStochastic, tspan, seed)
        P = P.T
        retry = False
        ## Extract Time points
        subset = P[gid, :][:, tps]
        df = pd.DataFrame(
            subset,
            index=pd.Index(genelist),
            columns=["E" + str(cellid) + "_" + str(i) for i in tps],
        )
        df.to_csv(outPrefix + "E" + str(cellid) + ".csv")
        ## Heuristic:
        ## If the largest value of a protein achieved in a simulation is
        ## less than 10% of the y_max, drop the simulation.
        ## This check stems from the observation that in some simulations,
        ## all genes go to the 0 steady state in some rare simulations.
        dfmax = df.max()
        for col in df.columns:
            colmax = df[col].max()
            if colmax < 0.1 * x_max:
                retry = True
                break

        if sampleCells:
            ## Write a single cell to file
            ## These samples allow for quickly and
            ## reproducibly testing the output.
            sampledf = utils.sampleCellFromTraj(
                cellid,
                tspan,
                P,
                varmapper,
                timeIndex,
                genelist,
                proteinlist,
                header,
                writeProtein=writeProtein,
            )
            sampledf = sampledf.T
            sampledf.to_csv(outPrefix + "E" + str(cellid) + "-cell.csv")

        trys += 1
        # write to file
        df.to_csv(outPrefix + "E" + str(cellid) + ".csv")

        if trys > 1:
            print("try", trys)



import numpy as np


def noise(x, t):
    # Controls noise proportional to
    # square root of activity
    c = 10.0  # 4.
    return c * np.sqrt(abs(x))


def deltaW(N, m, h, seed=0):
    """Generate sequence of Wiener increments for m independent Wiener
    processes W_j(t) j=0..m-1 for each of N time intervals of length h.
    From the sdeint implementation

    :returns:
        - dW : The [n, j] element has the value W_j((n+1)*h) - W_j(n*h) ( has shape (N, m) )
    """
    np.random.seed(seed)
    return np.random.normal(0.0, h, (N, m))


def eulersde(f, G, y0, tspan, pars, seed=0.0, dW=None):
    """
    Adapted from sdeint implementation https://github.com/mattja/sdeint/

    :param f: function defining ODE model. Should take vector of current state, current time, and list of parameter values as arguments.
    :type f: function
    :param pars: List of parameter values
    :type pars: list
    :param y0: list of initial values
    :type y0: list
    :param tspan: Array of timepoints to simulate
    :type tspan: ndarray
    :param seed: Seed to initialize random number generator
    :type seed: float
    :returns:
        - y: Array containing the time course of state variables
    """
    # From sdeint implementation
    N = len(tspan)
    h = (tspan[N - 1] - tspan[0]) / (N - 1)
    maxtime = tspan[-1]
    # allocate space for result
    d = len(y0)
    y = np.zeros((N + 1, d), dtype=type(y0[0]))

    if dW is None:
        # pre-generate Wiener increments (for d independent Wiener processes):
        dW = deltaW(N, d, h, seed=seed)
    y[0] = y0
    currtime = 0
    n = 0

    while currtime < maxtime:
        tn = currtime
        yn = y[n]
        dWn = dW[n, :]
        y[n + 1] = yn + f(yn, tn, pars) * h + np.multiply(G(yn, tn), dWn)
        # Ensure positive terms
        for i in range(len(y[n + 1])):
            if y[n + 1][i] < 0:
                y[n + 1][i] = yn[i]
        currtime += h
        n += 1
    return y


def simulateModel(Model, y0, parameters, isStochastic, tspan, seed):
    """Call numerical integration functions, either odeint() from Scipy,
    or simulator.eulersde() defined in simulator.py. By default, stochastic simulations are
    carried out using simulator.eulersde.

    :param Model: Function defining ODE model
    :type Model: function
    :param y0: array of initial values for each state variable
    :type y0: array
    :param parameters: list of parameter values to be used in simulations
    :type parameters: list
    :param isStochastic: User defined parameter. Default = True, can be turned off to perform ODE simulations.
    :type isStochastic: bool
    :param tspan: Time points to simulate
    :type tspan: ndarray
    :param seed: Seed to initialize random number generator
    :type seed: float
    :returns:
        - P: Time course from numerical integration
    :rtype: ndarray

    """
    if not isStochastic:
        P = odeint(Model, y0, tspan, args=(parameters,))
    else:
        P = eulersde(Model, noise, y0, tspan, parameters, seed=seed)
    return P


def getInitialCondition(
    ss,
    ModelSpec,
    rnaIndex,
    proteinIndex,
    genelist,
    proteinlist,
    varmapper,
    revvarmapper,
):
    """
    Calculate the initial values of all state variables.
    Takes into consideration user defined initial conditions, and computes the steady
    states of the protein variables based on the estimated values of their corresponding genes.

    :param ss: Steady state array
    :type ss: ndarray
    :param ModelSpec: Dictionary of dictionary specifying the ODE model, containing parameters, initial conditions and equations.
    :type ModelSpec: dict
    :param rnaIndex: list of indices of genes
    :type rnaIndex: list
    :param proteinIndex: List of indices of proteins
    :type proteinIndex: list
    :param genelist: List of names of all genes in the model
    :type genelist: list
    :param proteinlist: List of names of all proteins in the model
    :type proteinlist: list
    :param varmapper: Mapper: {variable name : index}
    :type varmapper: dict
    :param revvarmapper: Mapper: {index : variable name}
    :type revvarmapper: dict
    :returns:
        - newics: List containing new initial conditions
    """

    # Initialize
    new_ics = [0 for _ in range(len(varmapper.keys()))]
    # Set the mRNA ics
    for ind in rnaIndex:
        if ss[ind] < 0:
            ss[ind] = 0.0
        new_ics[ind] = ss[ind]
        if new_ics[ind] < 0:
            new_ics[ind] = 0
    for p in proteinlist:
        ind = revvarmapper["p_" + p]
        if ss[ind] < 0:
            ss[ind] = 0.0
        new_ics[ind] = ss[ind]
        if new_ics[ind] < 0:
            new_ics[ind] = 0
    # Calculate the Protein ics based on mRNA levels
    for genename in genelist:
        pss = (
            (ModelSpec["pars"]["r_" + genename])
            / (ModelSpec["pars"]["l_p_" + genename])
        ) * new_ics[revvarmapper["x_" + genename]]
        new_ics[revvarmapper["p_" + genename.replace("_", "")]] = pss
    return new_ics
